(ns com.aurellem.run.music
  (:use (com.aurellem.gb saves gb-driver util constants
                         items vbm characters money
                         rlm-assembly))
  (:use (com.aurellem.run util title save-corruption
                          bootstrap-0 bootstrap-1))
  (:import [com.aurellem.gb.gb_driver SaveState]))


(def music-base new-kernel)




(defn store [n address]
  (flatten
   [0xF5
    0xE5
    
    0x3E
    n
    
    0x21
    (reverse (disect-bytes-2 address))

    0x77
    
    0xE1
    0xF1]))

(defn infinite-loop []
  [0x18 0xFE])



(def divider-register 0xFF04)


(defrecord Bit-Note [frequency volume duration duty])

(defn clear-music-registers []
  (flatten
   [(store (Integer/parseInt "00000000" 2) 0xFF10)
    (store (Integer/parseInt "00000000" 2) 0xFF11)
    (store (Integer/parseInt "00000000" 2) 0xFF12)
    (store (Integer/parseInt "00000000" 2) 0xFF13)
    (store (Integer/parseInt "00000000" 2) 0xFF14)

    (store (Integer/parseInt "00000000" 2) 0xFF16) ;; pattern duty 000000
    (store (Integer/parseInt "00000000" 2) 0xFF17) ;; volume 0000
    (store (Integer/parseInt "00000000" 2) 0xFF18) ;; frequency-low
    (store (Integer/parseInt "00000000" 2) 0xFF19) ;; 00000 frequency-high

    (store (Integer/parseInt "00000000" 2) 0xFF1A)
    (store (Integer/parseInt "00000000" 2) 0xFF1B)
    (store (Integer/parseInt "00000000" 2) 0xFF1C)
    (store (Integer/parseInt "00000000" 2) 0xFF1D)
    (store (Integer/parseInt "00000000" 2) 0xFF1E)

    (store (Integer/parseInt "00000000" 2) 0xFF20)
    (store (Integer/parseInt "00000000" 2) 0xFF21)
    (store (Integer/parseInt "00000000" 2) 0xFF22)
    (store (Integer/parseInt "00000000" 2) 0xFF23)]))

(defn play-note
  "Play the note referenced by HL in the appropiate channel.
   Leaves desired-duration in A."
  []
  [0x2A   ;; load volume/frequency-high info
   0xF5   ;; push A
   0xE6
   (Integer/parseInt "11110000" 2) ;; volume mask
   0xE0
   0x17   ;; set volume
   0xF1   ;; pop A
   0xE6
   (Integer/parseInt "00000111" 2) ;; frequency-high mask
   0xE0   
   0x19   ;; set frequency-high
   
   0x2A   ;; load frequency low-bits
   0xE0
   0x18   ;; set frequency-low-bits

   0x7E   ;; load duration
   0x2B   ;; 
   0x2B   ;; HL-2 -> HL
   ]) 

(defn music-step []
  (flatten
   [(play-note)
    0xF5 ;; push A
    0xF0
    0x05 ;; load current ticks
    0xB8 ;; B holds previous sub-ticks, subtract it from A
    ;; if A-B caused a carry, then (B > A) is true, and
    ;; A = current-sub-tics, B = previous-sub-ticks, so
    ;; current-sub-ticks < previous-sub-ticks, which means that the
    ;; timer counter HAS overflowed.
    0x30 ;; increment C only if last result caused carry
    0x01
    0x0C

    0x47 ;; update sub-ticks (A->B)

    0xF1 ;; pop AF, now A contains desired-ticks

    0xB9 ;; compare with current ticks

    ;; if desired-ticks = current ticks
    ;;   go to next note ; set current set ticks to 0.

    0x20
    0x05

    0x23
    0x23
    0x23 ;; HL + 3 -> HL
    
    0x0E
    0x00])) ;; 0->C (current-ticks)

(defn test-timer []
  (flatten
   [0x3E
    0x01
    0xE0
    0x06 ;; set TMA to 0
    
    0x3E
    (Integer/parseInt "00000100" 2)
    0xE0
    0x07 ;; set TAC to 16384 Hz and activate timer
    
    (repeat
     500
     [0xF0
      0x05])]))


(defn music-kernel []
  (flatten
   [(clear-music-registers)

    0x21
    0x00
    0xD0 ;; set HL to 0xD000 == music-start
    0x0E
    0x00 ;; 0->C
    0x06
    0x00 ;; 0->B

    0x3E
    0x01
    0xE0
    0x06 ;; set TMA to 0
    
    0x3E
    (Integer/parseInt "00000110" 2)
    0xE0
    0x07 ;; set TAC to 65536 Hz and activate timer

    0xF0
    0x07
    
    (music-step)
    0x18
    (->signed-8-bit (+ (- (count (music-step)))
                       -2))]))

(def one-note
  [0xA0 0x00 0xFF])

(def many-notes
  (flatten (repeat 10 one-note)))

(def increasing-notes
  [0xA0 0x00 0x55
   0xA1 0x00 0x55
   0xA2 0x00 0x55
   0xA3 0x00 0x55
   0xA4 0x00 0x55
   0xA5 0x00 0x55
   0xA6 0x00 0x55
   0xA6 0x55 0xFF
   0xA6 0x55 0xFF
   0xA6 0x55 0xFF
   0x00 0x00 0xFF
   ])
   
   
(defn play-music [music-bytes]
  (let [program-target 0xC000
        music-target 0xD000]
    (-> (set-memory-range (second (music-base))
                          program-target (music-kernel))
        (set-memory-range music-target music-bytes)
        (PC! program-target))))


(defn test-note [music-bytes]
  (-> (set-memory-range (second (music-base))
                        0xC000 (concat (clear-music-registers)
                                       (play-note)
                                       (infinite-loop)))
      (set-memory-range 0xD000 music-bytes)
      (PC! 0xC000)
      (HL! 0xD000)
      ))


(defn run-program
  ([program]
     (let [target 0xC000]
       (-> (set-memory-range (second (music-base))
                             target program)
           (PC! target)))))

(defn trippy []
  (run-moves (play-music many-notes ) (repeat 8000 [])))

